# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import yaml

import fixtures
import testscenarios

import snapcraft
import integration_tests


class ManifestRecordingBaseTestCase(integration_tests.TestCase):
    """Test that the prime step records an annotated manifest.yaml

    The annotated file will be in prime/snap/manifest.yaml.

    """

    def setUp(self):
        super().setUp()
        self.useFixture(fixtures.EnvironmentVariable(
            'SNAPCRAFT_BUILD_INFO', '1'))


class ManifestRecordingTestCase(ManifestRecordingBaseTestCase):

    def test_prime_with_architectures(self):
        """Test the recorded manifest for a basic snap

        This snap doesn't have stage or build packages and is declared that it
        works on all architectures.
        """
        self.run_snapcraft('prime', project_dir='basic')

        recorded_yaml_path = os.path.join(
            self.prime_dir, 'snap', 'manifest.yaml')
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml.load(recorded_yaml_file)

        self.assertEqual(recorded_yaml['architectures'], ['all'])

    def test_prime_without_architectures_records_current_arch(self):
        """Test the recorded manifest for a basic snap

        This snap doesn't have stage or build packages and it is not declared
        that it works on all architectures, which makes it specific to the
        current architecture.
        """
        self.run_snapcraft('prime', project_dir='basic-without-arch')

        recorded_yaml_path = os.path.join(
            self.prime_dir, 'snap', 'manifest.yaml')
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml.load(recorded_yaml_file)

        self.assertEqual(
            recorded_yaml['architectures'],
            [snapcraft.ProjectOptions().deb_arch])


class ManifestRecordingBuildPackagesTestCase(
        testscenarios.WithScenarios, ManifestRecordingBaseTestCase):

    scenarios = [
        (snap, {'snap': snap}) for snap in
        ['build-package-global', 'build-packages-missing-dependency',
         'build-packages-without-dependencies']
    ]

    def test_prime_with_build_packages(self):
        """Test the recorded manifest for a snap with build packages

        This snap declares one global build package that has undeclared
        dependencies.

        """
        expected_packages = [
            'haskell-doc', 'haskell98-tutorial', 'haskell98-report']
        self.addCleanup(
            subprocess.call,
            ['sudo', 'apt', 'remove', '-y'] + expected_packages)

        self.run_snapcraft('prime', self.snap)

        expected_packages_with_version = [
            '{}={}'.format(
                package,
                integration_tests.get_package_version(
                    package, self.distro_series, self.deb_arch))
            for package in expected_packages
        ]

        recorded_yaml_path = os.path.join(
            self.prime_dir, 'snap', 'manifest.yaml')
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml.load(recorded_yaml_file)

        self.assertEqual(
            recorded_yaml['build-packages'], expected_packages_with_version)


class ManifestRecordingStagePackagesTestCase(ManifestRecordingBaseTestCase):

    def test_prime_records_packages_version(self):
        """Test the recorded manifest for a snap with packages

        This snap declares all the packages that it requires, there are
        no additional dependencies. The packages specify their version.
        """
        expected_packages = [
            'haskell-doc', 'haskell98-tutorial', 'haskell98-report']
        self.copy_project_to_cwd('stage-packages-without-dependencies')
        part_name = 'part-with-stage-packages'
        for package in expected_packages:
            self.set_stage_package_version(
                os.path.join('snap', 'snapcraft.yaml'), part_name, package)

        self.run_snapcraft('prime')

        with open(os.path.join('snap', 'snapcraft.yaml')) as source_yaml_file:
            source_yaml = yaml.load(source_yaml_file)

        recorded_yaml_path = os.path.join(
            self.prime_dir, 'snap', 'manifest.yaml')
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml.load(recorded_yaml_file)

        self.assertEqual(
            recorded_yaml['parts'][part_name]['stage-packages'],
            source_yaml['parts'][part_name]['stage-packages'])

    def test_prime_without_packages_version(self):
        """Test the recorded manifest for a snap with packages

        This snap declares all the packages that it requires, there are
        no additional dependencies. The packages don't specify their
        version.
        """
        self.run_snapcraft(
            'prime', project_dir='stage-packages-without-dependencies')

        with open(os.path.join('snap', 'snapcraft.yaml')) as source_yaml_file:
            source_yaml = yaml.load(source_yaml_file)
        part_name = 'part-with-stage-packages'
        expected_packages = [
            '{}={}'.format(
                package, integration_tests.get_package_version(
                    package, self.distro_series, self.deb_arch)) for
            package in source_yaml['parts'][part_name]['stage-packages']
        ]

        recorded_yaml_path = os.path.join(
            self.prime_dir, 'snap', 'manifest.yaml')
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml.load(recorded_yaml_file)

        self.assertEqual(
            recorded_yaml['parts'][part_name]['stage-packages'],
            expected_packages)

    def test_prime_with_packages_missing_dependency(self):
        """Test the recorded manifest for a snap with packages

        This snap declares one package that has undeclared dependencies.
        """
        self.copy_project_to_cwd('stage-packages-missing-dependency')
        part_name = 'part-with-stage-packages'
        self.set_stage_package_version(
            os.path.join('snap', 'snapcraft.yaml'),
            part_name, package='haskell-doc')
        self.run_snapcraft('prime')

        expected_packages = [
            '{}={}'.format(
                package,
                integration_tests.get_package_version(
                    package, self.distro_series, self.deb_arch))
            for package in [
                'haskell-doc', 'haskell98-tutorial', 'haskell98-report']
        ]

        recorded_yaml_path = os.path.join(
            self.prime_dir, 'snap', 'manifest.yaml')
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml.load(recorded_yaml_file)

        self.assertEqual(
            recorded_yaml['parts'][part_name]['stage-packages'],
            expected_packages)


class ManifestRecordingBzrSourceTestCase(
        integration_tests.BzrSourceBaseTestCase,
        ManifestRecordingBaseTestCase):

    def test_prime_with_bzr_source(self):
        self.copy_project_to_cwd('bzr-head')

        self.init_source_control()
        self.commit('"test-commit"', unchanged=True)

        self.run_snapcraft('prime')

        recorded_yaml_path = os.path.join(
            self.prime_dir, 'snap', 'manifest.yaml')
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml.load(recorded_yaml_file)

        commit = self.get_revno()
        self.assertEqual(
            recorded_yaml['parts']['bzr']['source-commit'], commit)


class ManifestRecordingGitSourceTestCase(
        integration_tests.GitSourceBaseTestCase,
        ManifestRecordingBaseTestCase):

    def test_prime_with_git_source(self):
        self.copy_project_to_cwd('git-head')

        self.init_source_control()
        self.commit('"test-commit"', allow_empty=True)

        self.run_snapcraft('prime')

        recorded_yaml_path = os.path.join(
            self.prime_dir, 'snap', 'manifest.yaml')
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml.load(recorded_yaml_file)

        commit = self.get_revno()
        self.assertEqual(
            recorded_yaml['parts']['git']['source-commit'], commit)


class ManifestRecordingHgSourceTestCase(
        integration_tests.HgSourceBaseTestCase,
        ManifestRecordingBaseTestCase):

    def test_prime_with_hg_source(self):
        self.copy_project_to_cwd('hg-head')

        self.init_source_control()
        open('1', 'w').close()
        self.commit('1', '1')

        self.run_snapcraft('prime')

        recorded_yaml_path = os.path.join(
            self.prime_dir, 'snap', 'manifest.yaml')
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml.load(recorded_yaml_file)

        commit = self.get_id()
        self.assertEqual(
            recorded_yaml['parts']['mercurial']['source-commit'], commit)


class ManifestRecordingSubversionSourceTestCase(
        integration_tests.SubversionSourceBaseTestCase,
        ManifestRecordingBaseTestCase):

    def test_prime_with_subversion_source(self):
        self.copy_project_to_cwd('svn-pull')

        self.init_source_control()
        self.checkout(
            'file:///{}'.format(os.path.join(self.path, 'repo')), 'local')

        open(os.path.join('local', 'file'), 'w').close()
        self.add('file', cwd='local')
        self.commit('test', cwd='local/')
        self.update(cwd='local/')
        subprocess.check_call(
            ['rm', '-rf', 'local/'], stdout=subprocess.DEVNULL)

        self.run_snapcraft('prime')

        recorded_yaml_path = os.path.join(
            self.prime_dir, 'snap', 'manifest.yaml')
        with open(recorded_yaml_path) as recorded_yaml_file:
            recorded_yaml = yaml.load(recorded_yaml_file)

        self.assertEqual(
            recorded_yaml['parts']['svn']['source-commit'], '1')
